/*
    Turbo BDM Light ColdFire
    Copyright (C) 2006  Daniel Malik

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <windows.h>
#include "tblcf_usb.h"
#include "log.h"
#include "tblcf_hwdesc.h"
#include "libusb.h"

static struct usb_device *usb_devs[10]={NULL};	/* array pointing to all Turbo BDM Light ColdFire devices found terminated by NULL pointer */
static usb_dev_handle *usb_devh=NULL;

/* provides low level USB functions which talk to the hardware */

/* initialisation */
void tblcf_usb_init(void) {
	usb_init();				/* init LIBUSB */
	usb_set_debug(0);		/* set debug level to minimum */
}

/* find all TBLCF devices attached to the computer */
void tblcf_usb_find_devices(unsigned short int product_id) {
	struct usb_bus *libusb_bus;
	struct usb_device *libusb_dev;
  	struct usb_version *version;
	unsigned int i=0;
	usb_devs[0]=NULL;				/* terminate the list */
	version = usb_get_version();
	if (version==NULL) {
		/* something quite wrong... */
		return;
	}
    print("USBLIB DLL version: %i.%i.%i.%i\r\n",version->dll.major,version->dll.minor,version->dll.micro,version->dll.nano);
	if (version->driver.major==-1) {
		/* driver not running! */
		print("USB: Driver not running\r\n");
		return;
	}
	print("USBLIB Driver version: %i.%i.%i.%i\r\n",version->driver.major, version->driver.minor,version->driver.micro, version->driver.nano);
	usb_find_busses();		/* enumerate all busses */
	usb_find_devices();		/* enumerate all devices */
	for (libusb_bus = usb_get_busses(); libusb_bus; libusb_bus = libusb_bus->next) {		/* scan through all busses */
    	for (libusb_dev = libusb_bus->devices; libusb_dev; libusb_dev = libusb_dev->next) {	/* scan through all devices */
  			if ((libusb_dev->descriptor.idVendor==TBLCF_VID)&&(libusb_dev->descriptor.idProduct==product_id)) {
				/* found a device */
				usb_devs[i++]=libusb_dev;	/* add pointer to the device */
				usb_devs[i]=NULL;			/* terminate the list again */
				if (i>=(sizeof(usb_devs)/sizeof(struct usb_device*))) return;	/* too many devices! */
			}
		}
	}
}

/* returns number of attached TBDML devices */
unsigned int tblcf_usb_cnt(void) {
	unsigned int i=0;
	while ((i<(sizeof(usb_devs)/sizeof(struct usb_device*)))&&(usb_devs[i]!=NULL)) i++;	/* count the entries */
	return(i);
}

/* open connection to device enumerated by tblcf_usb_find_devices */
/* returns 0 on success and 1 on error */
unsigned char tblcf_usb_open(unsigned int device_no) {
	usb_devh = usb_open(usb_devs[device_no]);
	if (usb_devh==NULL) return (1);
	print("USB Device open\r\n");
	if (usb_set_configuration(usb_devh,1)) return(1);	/* TBLCF has only one valid configuration */
	print("USB Configuration set\r\n");
	if (usb_claim_interface(usb_devh,0)) return(1);		/* TBLCF has only 1 interface */
	print("USB Interface claimed\r\n");
	return (0);
}

/* closes connection to the currently open device */
void tblcf_usb_close(void) {
	if (usb_devh!=NULL) {
		usb_release_interface(usb_devh,0);			/* release the interface */
		print("USB Interface released\r\n");
		usb_close(usb_devh);						/* close the device */
		print("USB Device closed\r\n");
		usb_devh=NULL;								/* indicate that no device is open */
	}
}

/* Message data format:
				1 byte: size of cmd+data
			    1 byte: cmd
				(size-1) bytes: data
*/

/* sends a message to the TBDML device over EP0 */
/* returns 0 on success and 1 on error */
/* since the EP0 transfer is unidirectional in this case, data returned by the device must be read separately */
unsigned char tblcf_usb_send_ep0(unsigned char * data) {
	unsigned char * count = data;		/* data count is the first byte of the message */
 	int i;
	if (usb_devh==NULL) {
		print("USB EP0 send: device not open\r\n");
		return(1);
	}
	print("USB EP0 send:\r\n");
	print_dump(data,(*count)+1);
	i=usb_control_msg(usb_devh, 0x40, *(data+1), (*(data+2))+256*(*(data+3)), (*(data+4))+256*(*(data+5)), data+6, ((*count)>5)?((*count)-5):0, TIMEOUT);
	if (i<0) return(1); else return(0);
}

/* sends a message to the TBDML device over EP0 which instruct the device to exec command and return data */
/* returns 0 on success and 1 on error */
/* data count in the message is number of bytes EXPECTED/REQUIRED from the device */
/* the device will get the cmd number and 4 following data bytes */
unsigned char tblcf_usb_recv_ep0(unsigned char * data) {
	unsigned char count = *data;		/* data count is the first byte of the message */
 	int i;
	if (usb_devh==NULL) {
		print("USB EP0 receive request: device not open\r\n");
		return(1);
	}
	print("USB EP0 receive request:\r\n");
	print_dump(data,6);
	i=usb_control_msg(usb_devh, 0xC0, *(data+1), (*(data+2))+256*(*(data+3)), (*(data+4))+256*(*(data+5)), data, count, TIMEOUT);
	print("USB EP0 receive:\r\n");
	print_dump(data,count);
	if (i<0) return(1); else return(0);
}

/* sends a message to the TBDML device over EP2 */
/* returns 0 on success and 1 on error */
/* data the device wants to return need to be read out in separate recv transaction */
unsigned char tblcf_usb_send_ep2(unsigned char *data) {
	unsigned char * count = data;		/* data count is the first byte of the message */
 	int i;
	if (usb_devh==NULL) {
		print("USB EP2 send: device not open\r\n");
		return(1);
	}
	print("USB EP2 send:\r\n");
	print_dump(data,(*count)+1);
	i=usb_bulk_write(usb_devh, 0x02, data, (*count)+1, TIMEOUT);
	if (i<0) return(1); else return(0);
}

/* receives data from EP2 */
/* returns 0 on success and 1 on error */
/* data count in the message is number of bytes EXPECTED/REQUIRED from the device */
unsigned char tblcf_usb_recv_ep2(unsigned char * data) {
	unsigned char count = *data;		/* data count is the first byte of the message */
 	int i;
	if (usb_devh==NULL) {
		print("USB EP2 receive request: device not open\r\n");
		return(1);
	}
	print("USB EP2 receive request:\r\n");
	print_dump(data,6);
	i=usb_bulk_read(usb_devh, 0x82, data, count, TIMEOUT);
	print("USB EP2 receive (%d byte(s)):\r\n",i);
	print_dump(data,count);
	if (i<0) return(1); else return(0);
}

/* routines to interact with the ICP code on JB16 */

/* programs given number of bytes into flash of the device from given address */
/* returns 0 on succes, -1 on USB error and 1 on compare error */
char icp_program(unsigned char * data, unsigned int address, unsigned int count) {
	int i;
	unsigned char result;
	if (usb_devh==NULL) {
		print("ICP program: device not open\r\n");
		return(-1);
	}
	while (count>ICP_MAX_PACKET_SIZE) {
		print("ICP program %d byte from address 0x%04X:\r\n",ICP_MAX_PACKET_SIZE,address);
		print_dump(data,ICP_MAX_PACKET_SIZE);
		i=usb_control_msg(usb_devh, 0x40, ICP_PROGRAM, address, address+ICP_MAX_PACKET_SIZE-1, data, ICP_MAX_PACKET_SIZE, TIMEOUT);
		if (i<0) {
			print("ICP program: request failed\r\n");
			return(-1);
		}
		sleep(10);	/* give the part plenty of time to do the programming */
		do {
			i=usb_control_msg(usb_devh, 0xC0, ICP_GET_RESULT, 0, 0, &result, 1, TIMEOUT);
			if (i<0) {
				print("ICP program: get result request failed\r\n");
				return(-1);
			}
		} while (result==ICP_RESULT_BUSY);	/* keep checking the result while busy */
		if (result!=ICP_RESULT_OK) {
			print("ICP programming error (%02X)\r\n",result);
			return(1);
		}
		count-=ICP_MAX_PACKET_SIZE;
		address+=ICP_MAX_PACKET_SIZE;
		data+=ICP_MAX_PACKET_SIZE;
	}
	if (count) {
		print("ICP program %d byte(s) from address 0x%04X:\r\n",count,address);
		print_dump(data,count);
		i=usb_control_msg(usb_devh, 0x40, ICP_PROGRAM, address, address+count-1, data, count, TIMEOUT);
		if (i<0) {
			print("ICP program: request failed\r\n");
			return(-1);
		}
		sleep(10);	/* give the part plenty of time to do the programming */
		do {
			i=usb_control_msg(usb_devh, 0xC0, ICP_GET_RESULT, 0, 0, &result, 1, TIMEOUT);
			if (i<0) {
				print("ICP program: get result request failed\r\n");
				return(-1);
			}
		} while (result==ICP_RESULT_BUSY);	/* keep checking the result while busy */
		if (result!=ICP_RESULT_OK) {
			print("ICP programming error (%02X)\r\n",result);
			return(1);
		}
	}
	return(0);
}

/* performs mass erase of device flash */
/* returns 0 on success , -1 on USB error and 1 on operation error */
char icp_mass_erase(void) {
	int i;
	unsigned char result;
	if (usb_devh==NULL) {
		print("ICP mass erase: device not open\r\n");
		return(-1);
	}
	print("ICP mass erase\r\n");
	i=usb_control_msg(usb_devh, 0x40, ICP_MASS_ERASE, 0, 0, NULL, 0, TIMEOUT);
	if (i<0) {
		print("ICP mass erase request failed\r\n");
		return(-1);
	}
	Sleep(400);	/* wait 250ms (time of mass erase) + lots of extra to make sure the operation is finished by the time new traffic apears on the USB bus */
	do {
		i=usb_control_msg(usb_devh, 0xC0, ICP_GET_RESULT, 0, 0, &result, 1, TIMEOUT);
		if (i<0) {
			print("ICP mass erase: get result request failed\r\n");
			return(-1);
		}
	} while (result==ICP_RESULT_BUSY);	/* keep checking the result while busy */
	if (result!=ICP_RESULT_OK) {
		print("ICP mass erase error (0x%02X)\r\n",result);
		return(1);
	}
	return(0);
}

/* performs block erase on block containing given address */
/* returns 0 on success, -1 on USB error and 1 on operation error */
char icp_block_erase(unsigned int address) {
	int i;
	unsigned char result;
	unsigned char blank_test_array[ICP_FLASH_BLOCK_SIZE];
	if (usb_devh==NULL) {
		print("ICP block erase: device not open\r\n");
		return(-1);
	}
	address&=(65536-ICP_FLASH_BLOCK_SIZE);	/* make sure the address is only 16 bit and align it to the closest lower block boundary */
	print("ICP block erase\r\n");
	for (i=0;i<ICP_FLASH_BLOCK_SIZE;i++) blank_test_array[i]=0xff;
	i=icp_verify(blank_test_array, address, ICP_FLASH_BLOCK_SIZE);
	if (i<0) {
		print("ICP block erase: verify request failed\r\n");
		return(-1);
	}
	if (i==0) return(0);	/* block is already erased */
	/* sector is not blank, must perform block erase */
	print("ICP block erase: block not empty, erasing...\r\n");
	i=usb_control_msg(usb_devh, 0x40, ICP_BLOCK_ERASE, address, address+ICP_FLASH_BLOCK_SIZE-1, NULL, 0, TIMEOUT);
	if (i<0) {
		print("ICP block erase request failed\r\n");
		return(-1);
	}
	Sleep(30);	/* wait 10ms (time of block erase) + lots of extra to make sure the operation is finished by the time new traffic apears on the USB bus */
	do {
		i=usb_control_msg(usb_devh, 0xC0, ICP_GET_RESULT, 0, 0, &result, 1, TIMEOUT);
		if (i<0) {
			print("ICP block erase: get result request failed\r\n");
			return(-1);
		}
	} while (result==ICP_RESULT_BUSY);	/* keep checking the result while busy */
	if (result!=ICP_RESULT_OK) {
		print("ICP block erase error (0x%02X)\r\n",result);
		return(1);
	}
	return(0);
}

/* verifies given number of bytes against contents of flash of the device from given address */
/* returns 0 on succesful compare, -1 on USB error and 1 on compare error */
char icp_verify(unsigned char * data, unsigned int address, unsigned int count) {
	int i;
	unsigned char result;
	if (usb_devh==NULL) {
		print("ICP verify: device not open\r\n");
		return(-1);
	}
	while (count>ICP_MAX_PACKET_SIZE) {
		print("ICP verify %d bytes from address 0x%04X:\r\n",ICP_MAX_PACKET_SIZE,address);
		print_dump(data,ICP_MAX_PACKET_SIZE);
		i=usb_control_msg(usb_devh, 0x40, ICP_VERIFY, address, address+ICP_MAX_PACKET_SIZE-1, data, ICP_MAX_PACKET_SIZE, TIMEOUT);
		if (i<0) {
			print("ICP verify request failed\r\n");
			return(-1);
		}
		Sleep(5);	/* wait to make sure the operation is finished by the time new traffic apears on the USB bus */
		do {
			i=usb_control_msg(usb_devh, 0xC0, ICP_GET_RESULT, 0, 0, &result, 1, TIMEOUT);
			if (i<0) {
				print("ICP verify: get result request failed\r\n");
				return(-1);
			}
		} while (result==ICP_RESULT_BUSY);	/* keep checking the result while busy */
		if (result!=ICP_RESULT_OK) {
			print("ICP verification error\r\n");
			return(1);
		}
		count-=ICP_MAX_PACKET_SIZE;
		address+=ICP_MAX_PACKET_SIZE;
		data+=ICP_MAX_PACKET_SIZE;
	}
	if (count) {
		print("ICP verify %d byte(s) from address 0x%04X:\r\n",count,address);
		print_dump(data,count);
		i=usb_control_msg(usb_devh, 0x40, ICP_VERIFY, address, address+count-1, data, count, TIMEOUT);
		if (i<0) {
			print("ICP verify request failed\r\n");
			return(-1);
		}
		Sleep(5);	/* wait to make sure the operation is finished by the time new traffic apears on the USB bus */
		do {
			i=usb_control_msg(usb_devh, 0xC0, ICP_GET_RESULT, 0, 0, &result, 1, TIMEOUT);
			if (i<0) {
				print("ICP verify: get result request failed\r\n");
				return(-1);
			}
		} while (result==ICP_RESULT_BUSY);	/* keep checking the result while busy */
		if (result!=ICP_RESULT_OK) {
			print("ICP verification error\r\n");
			return(1);
		}
	}
	return(0);
}
